/*

	PROJECT:		m0d_s0beit_sa
	LICENSE:		See LICENSE in the top level directory
	COPYRIGHT:		Copyright 2007, 2008, 2009, 2010 we_sux

	m0d_s0beit_sa is available from http://code.google.com/p/m0d-s0beit-sa/

	m0d_s0beit_sa is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	m0d_s0beit_sa is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with m0d_s0beit_sa.  If not, see <http://www.gnu.org/licenses/>.

*/
#include "main.h"

// new function to jump into vehicles without jacking (also for single player)
void vehicleJumper ( int iVehicleID )
{
	// can't touch this
	if ( iVehicleID == VEHICLE_SELF )
		return;

	// get vehicle_info
	struct vehicle_info *pVehicle = vehicle_info_get( iVehicleID, 0 );

	// check that the vehicle is legit
	if ( isBadPtr_GTA_pVehicleInfo(pVehicle) )
		return;

	traceLastFunc( "vehicleJumper()" );

	// if SAMP is loaded, check if the vehicle is streamed in
	if ( g_SAMP != NULL )
	{
		int iVehicleSAMPID = getSAMPVehicleIDFromGTAVehicle( pVehicle );
		if ( isBadPtr_SAMP_iVehicleID(iVehicleSAMPID) )
			return;
	}

	if ( pVehicle->hitpoints == 0.0f )
	{
		cheat_state_text( "Vehicle is destroyed" );
		return;
	}

	if ( cheat_state->actor.air_brake )
	{
		cheat_state_text( "On foot airbrake must be disabled" );
		return;
	}

	if ( cheat_state->actor.stick )
	{
		cheat_state_text( "On foot stick must be disabled" );
		return;
	}

	if ( !pVehicle->base.bIsVisible )
	{
		cheat_state_text( "Vehicle is not visible." );
		return;
	}

	struct actor_info	*self = actor_info_get( ACTOR_SELF, 0 );
	if ( self != NULL && pVehicle->base.interior_id != self->base.interior_id )
	{
		cheat_state_text( "Vehicle is in another interior." );
		return;
	}

	int iGTAVehicleID;
	iGTAVehicleID = ScriptCarId( pVehicle );

	if ( pVehicle->passengers[0] == self )
		return;

	if ( !pVehicle->base.bUsesCollision )
	{
		//cheat_state_text("Can't get in a vehicle that doesn't have collisions enabled.");
		cheat_state_text( "I do as you command." );
		cheat_state->_generic.nocols_enabled = 0;
		cheat_state->_generic.nocols_change_tick = GetTickCount();
		pVehicle->base.bUsesCollision = 1;
	}

	// put into first available seat
	if ( pVehicle->passengers[0] == NULL )
	{
		ScriptCommand( &put_actor_in_car, ScriptActorId(self), iGTAVehicleID );
		ScriptCommand( &restore_camera_with_jumpcut );
		ScriptCommand( &set_camera_directly_behind );
		ScriptCommand( &restore_camera_with_jumpcut );
		return;
	}

	const int	seat_count = gta_vehicle_get_by_id( pVehicle->base.model_alt_id )->passengers;
	if ( seat_count > 0 )
	{
		for ( int seat = 1; seat <= seat_count; seat++ )
		{
			if ( pVehicle->passengers[seat] == NULL )
			{
				ScriptCommand( &put_actor_in_car_passenger, ScriptActorId(self), iGTAVehicleID, seat - 1 );
				ScriptCommand( &restore_camera_with_jumpcut );
				ScriptCommand( &set_camera_directly_behind );
				ScriptCommand( &restore_camera_with_jumpcut );
				return;
			}
		}
	}

	// no seats left, oh well
	cheat_state_text( "No seats left to teleport into." );
	return;
}

void cheat_vehicle_teleport ( struct vehicle_info *info, const float pos[3], int interior_id )
{
	if ( info == NULL )
		return;

	traceLastFunc( "cheat_vehicle_teleport()" );

	float				diff[3];
	float				new_pos[3];
	struct vehicle_info *temp;

	vect3_vect3_sub( pos, &info->base.matrix[4 * 3], diff );

	for ( temp = info; temp != NULL; temp = temp->trailer )
	{
		vect3_vect3_add( &temp->base.matrix[4 * 3], diff, new_pos );

		vehicle_detachables_teleport( temp, &temp->base.matrix[4 * 3], new_pos );
		vect3_copy( new_pos, &temp->base.matrix[4 * 3] );

		vect3_zero( temp->speed );
		vect3_zero( temp->spin );

		gta_interior_id_set( interior_id );
		temp->base.interior_id = ( uint8_t ) interior_id;

		if ( !set.trailer_support )
			break;
	}
}

void cheat_handle_vehicle_unjacker ( struct vehicle_info *info, float time_diff )
{
	//ds
}

//Making the vehicle instant jumping and the trailer attaching easier
void cheat_handle_vehicle_nocols ( struct vehicle_info *info )
{
	if ( isBadPtr_GTA_pVehicleInfo(info) )
		return;

	traceLastFunc( "cheat_handle_vehicle_nocols()" );

	if ( cheat_state->_generic.nocols_toggled )
	{
		cheat_state->_generic.nocols_enabled = 1;
	}
	else if ( cheat_state->_generic.nocols_enabled )
	{
		//for dumb_menu vehicle jumping and trailer attaching
		int					veh_id = vehicle_find_nearest( NULL );
		struct vehicle_info *veh = vehicle_info_get( veh_id, 0 );
		float				dist[3] = { 0.0f, 0.0f, 0.0f };
		if ( veh != NULL )
			vect3_vect3_sub( &info->base.matrix[4 * 3], &veh->base.matrix[4 * 3], dist );
		if ( (info->trailer != NULL && (GetTickCount() - 500) > cheat_state->_generic.nocols_change_tick)
		 ||	 (info->trailer == NULL && vect3_length(dist) >= 10.0f) ) cheat_state->_generic.nocols_enabled = 0;
	}

	return;
}

void cheat_handle_vehicle_unflip ( struct vehicle_info *info, float time_diff )
{
	traceLastFunc( "cheat_handle_vehicle_unflip()" );

	/* Unflip */
	//if ( KEY_PRESSED(set.key_unflip) )
	//{ }
	/* Rotate vehicle while the unflip key is held down */
	if ( KEY_DOWN(set.key_unflip) )
	{
		traceLastFunc( "cheat_handle_vehicle_unflip()" );

		// inter-frame timing info from the game
		float		fTimeStep = *(float *)0xB7CB5C;

		// get our vehicle, gravity, and matrix
		CVehicle	*cveh = getSelfCVehicle();
		CVector		cvehGrav;
		CMatrix		cvehMatrix;
		cveh->GetGravity( &cvehGrav );
		cveh->GetMatrix( &cvehMatrix );

		// get "down" from vehicle model
		CVector rotationAxis = cheat_vehicle_getPositionUnder( cveh );

		// normalize our vectors
		cvehGrav.Normalize();
		rotationAxis.Normalize();

		// axis and rotation for gravity
		float	theta = acos( rotationAxis.DotProduct(&cvehGrav) );
		if ( theta > FLOAT_EPSILON )
		{
			rotationAxis.CrossProduct( &cvehGrav );
			rotationAxis.Normalize();
			rotationAxis.ZeroNearZero();
			cvehMatrix = cvehMatrix.Rotate( &rotationAxis, -theta );
		}

		// slow turn to the right
		theta = M_PI / ( 75.0f / fTimeStep );

		CVector slowTurnAxis = cvehMatrix.vUp;
		slowTurnAxis.Normalize();
		slowTurnAxis.ZeroNearZero();
		cvehMatrix = cvehMatrix.Rotate( &cvehMatrix.vUp, theta );

		// set the new matrix
		cveh->SetMatrix( &cvehMatrix );

		// no more spin cycle for you
		CVector vZero ( 0.0f, 0.0f, 0.0f );
		cveh->SetTurnSpeed( &vZero );
	}
}

void cheat_vehicle_air_brake_set ( int enabled )
{
	cheat_state->vehicle.air_brake = enabled;
	cheat_handle_vehicle_air_brake( NULL, 0.0f );
}

void cheat_handle_vehicle_air_brake ( struct vehicle_info *info, double time_diff )
{
	static float		orig_matrix[16];
	static int			orig_matrix_set;
	struct vehicle_info *temp;

	if ( info == NULL )
	{
		orig_matrix_set = 0;
		return;
	}

	traceLastFunc( "cheat_handle_vehicle_air_brake()" );

	// handle key presses
	if ( set.air_brake_toggle )
	{
		if ( KEY_PRESSED(set.key_air_brake_mod) )
			cheat_state->vehicle.air_brake ^= 1;

		if ( KEY_PRESSED(set.key_air_brake_mod2) && cheat_state->vehicle.air_brake )
			cheat_state->vehicle.air_brake_slowmo ^= 1;
	}
	else
	{
		if ( KEY_PRESSED(set.key_air_brake_mod) )
			cheat_state->vehicle.air_brake = 1;
		else if ( KEY_RELEASED(set.key_air_brake_mod) )
			cheat_state->vehicle.air_brake = 0;

		if ( KEY_PRESSED(set.key_air_brake_mod2) && cheat_state->vehicle.air_brake )
			cheat_state->vehicle.air_brake_slowmo = 1;
		else if ( KEY_RELEASED(set.key_air_brake_mod2) && cheat_state->vehicle.air_brake )
			cheat_state->vehicle.air_brake_slowmo = 0;
	}

	// do airbrake
	if ( !cheat_state->vehicle.air_brake )
	{
		orig_matrix_set = 0;
		cheat_state->vehicle.air_brake_slowmo = 0;
	}
	else
	{
		if ( !orig_matrix_set )
		{
			matrix_copy( info->base.matrix, orig_matrix );
			orig_matrix_set = 1;
		}

		if ( !KEY_DOWN(set.key_unflip) )	/* allow the unflip rotation feature to work */
			matrix_copy( orig_matrix, info->base.matrix );

		/* XXX allow some movement */
		for ( temp = info; temp != NULL; temp = temp->trailer )
		{
			/* prevent all kinds of movement */
			vect3_zero( temp->speed );
			vect3_zero( temp->spin );

			if ( !set.trailer_support )
				break;
		}

		static uint32_t time_start;
		float			*matrix = info->base.matrix;
		float			d[4] = { 0.0f, 0.0f, 0.0f, time_diff * set.air_brake_speed };
		float			xyvect[4] = { 0.0f, 0.0f, 0.0f, 0.0f }; /* rotation vector */
		float			zvect[4] = { 0.0f, 0.0f, 0.0f, 0.0f };	/* rotation vector */
		float			theta = set.air_brake_rot_speed * time_diff * M_PI * 2.0f;

		if ( cheat_state->vehicle.air_brake_slowmo )
		{
			d[3] /= 10.0f;
			theta /= 2.0f;
		}

		if ( KEY_DOWN(set.key_air_brake_forward) )
			d[0] += 1.0f;
		if ( KEY_DOWN(set.key_air_brake_backward) )
			d[0] -= 1.0f;
		if ( KEY_DOWN(set.key_air_brake_left) )
			d[1] += 1.0f;
		if ( KEY_DOWN(set.key_air_brake_right) )
			d[1] -= 1.0f;
		if ( KEY_DOWN(set.key_air_brake_up) )
			d[2] += 1.0f;
		if ( KEY_DOWN(set.key_air_brake_down) )
			d[2] -= 1.0f;

		if ( !near_zero(set.air_brake_accel_time) )
		{
			if ( !vect3_near_zero(d) )
				time_start = ( time_start == 0 ) ? time_get() : time_start;
			else
				time_start = 0; /* no keys pressed */

			/* acceleration */
			if ( time_start != 0 )
			{
				float	t = TIME_TO_FLOAT( time_get() - time_start );
				if ( t < set.air_brake_accel_time )
					d[3] *= t / set.air_brake_accel_time;
			}
		}

		/* pitch (x-axis) */
		if ( KEY_DOWN(set.key_air_brake_rot_pitch1) )
			xyvect[0] += 1.0f;
		if ( KEY_DOWN(set.key_air_brake_rot_pitch2) )
			xyvect[0] -= 1.0f;

		/* roll (y-axis) */
		if ( KEY_DOWN(set.key_air_brake_rot_roll1) )
			xyvect[1] += 1.0f;
		if ( KEY_DOWN(set.key_air_brake_rot_roll2) )
			xyvect[1] -= 1.0f;

		/* yaw (z-axis) */
		if ( KEY_DOWN(set.key_air_brake_rot_yaw1) )
			zvect[2] -= 1.0f;
		if ( KEY_DOWN(set.key_air_brake_rot_yaw2) )
			zvect[2] += 1.0f;

		if ( !vect3_near_zero(xyvect) )
		{
			vect3_normalize( xyvect, xyvect );

			matrix_vect4_mult( matrix, xyvect, xyvect );
			matrix_vect3_rotate( matrix, xyvect, theta, matrix );
		}

		if ( !vect3_near_zero(zvect) )
		{
			vect3_normalize( zvect, zvect );

			matrix_vect3_rotate( matrix, zvect, theta, matrix );
		}

		switch ( set.air_brake_behaviour )
		{
		case 0:
			matrix[4 * 1 + 0] += d[0] * d[3];
			matrix[4 * 1 + 1] += d[1] * d[3];
			matrix[4 * 1 + 2] += d[2] * d[3];
			break;

		case 1: /* mixed 2d/3d mode */
			{
				/* convert the vehicle's 3d rotation vector to a 2d angle */
				float	a = atan2f( matrix[4 * 1 + 0], matrix[4 * 1 + 1] );

				/* calculate new position */
				matrix[4 * 3 + 0] += ( d[0] * sinf(a) - d[1] * cosf(a) ) * d[3];
				matrix[4 * 3 + 1] += ( d[0] * cosf(a) + d[1] * sinf(a) ) * d[3];
				matrix[4 * 3 + 2] += d[2] * d[3];
			}
			break;

		case 2: /* full 3d movement */
			if ( !vect3_near_zero(d) )
			{
				float	vect[4] = { -d[1], d[0], d[2], 0.0f };	/* swap x & y + invert y */
				float	out[4];

				/* out = matrix * norm(d) */
				vect3_normalize( vect, vect );
				matrix_vect4_mult( matrix, vect, out );

				matrix[4 * 3 + 0] += out[0] * d[3];
				matrix[4 * 3 + 1] += out[1] * d[3];
				matrix[4 * 3 + 2] += out[2] * d[3];
			}
			break;
		}

		matrix_copy( matrix, orig_matrix );

		// heh
		int		gonadsMult = 1000;
		float	strifeMult = 0.000001f;
		int		gonads = rand() % gonadsMult;
		float	strife = (double)gonads * strifeMult;
		if ( strife < strifeMult * gonadsMult / 2 )
			strife -= strifeMult * gonadsMult;
		info->m_SpeedVec.fX = strife;
		gonads = rand() % gonadsMult;
		strife = (double)gonads * strifeMult;
		if ( strife < strifeMult * gonadsMult / 2 )
			strife -= strifeMult * gonadsMult;
		info->m_SpeedVec.fY = strife;
		strifeMult = 0.000000001f;
		gonads = rand() % gonadsMult;
		strife = (double)gonads * strifeMult;
		if ( strife < strifeMult * gonadsMult / 2 )
			strife -= strifeMult * gonadsMult;
		info->m_SpeedVec.fZ = strife;
	}
}

void cheat_handle_vehicle_warp ( struct vehicle_info *info, float time_diff )
{
	traceLastFunc( "cheat_handle_vehicle_warp()" );

	static struct vehicle_state state;
	static float				raildist;
	static uint32_t				warp_time = 0;
	static int					warping = 0;

	if ( warp_time != 0 )
	{
		if ( time_get() - warp_time > MSEC_TO_TIME(100) )
		{
			warp_time = 0;
		}
		else
		{
			vehicle_state_restore( info, &state );
			if ( info->vehicle_type == VEHICLE_TYPE_TRAIN )
				info->m_fTrainRailDistance = raildist;
		}
	}

	/* warp factor: fuck you */
	if ( KEY_PRESSED(set.key_warp_mod) )
	{
		struct vehicle_info *temp;
		float				dir[4] = { 0.0f, 1.0f, 0.0f, 0.0f };
		float				vect[4];

		if ( set.warp_use_speed && !vect3_near_zero(info->speed) && vect3_length(info->speed) >= 0.01f )
		{
			/* use normalized speed vector */
			vect3_copy( info->speed, dir );
			vect3_normalize( dir, vect );
		}
		else
		{
			/* use vehicle direction matrix */
			matrix_vect4_mult( info->base.matrix, dir, vect );
			if ( vect3_near_zero(vect) )
				return; /* shouldn't happen */
		}

		vehicle_state_store( info, &state );
		if ( info->vehicle_type == VEHICLE_TYPE_TRAIN )
			raildist = info->m_fTrainRailDistance;
		warp_time = 0;
		for ( temp = info; temp != NULL; temp = temp->trailer )
		{
			vect3_mult( vect, set.warp_speed, temp->speed );

			if ( !set.trailer_support )
				break;
		}
	}

	if ( KEY_DOWN(set.key_warp_mod) && info->vehicle_type == VEHICLE_TYPE_TRAIN )
		info->m_fTrainSpeed = set.warp_speed;

	if ( KEY_RELEASED(set.key_warp_mod) )
	{
		struct vehicle_info *temp;

		warp_time = time_get();
		vehicle_state_restore( info, &state );
		if ( info->vehicle_type == VEHICLE_TYPE_TRAIN )
			info->m_fTrainRailDistance = raildist;
		for ( temp = info; temp != NULL; temp = temp->trailer )
		{
			temp->collision_something = 0.0f;

			if ( !set.trailer_support )
				break;
		}
	}
}

void cheat_handle_vehicle_hop ( struct vehicle_info *info, float time_diff )
{
	traceLastFunc( "cheat_handle_vehicle_hop()" );

	if ( KEY_DOWN(set.key_vehicle_hop) )
	{
		CVehicle	*cveh = pGameInterface->GetPools()->GetVehicle( (DWORD *)info );
		CVector		vecSpeedAdd, vecMoveSpeed;
		float		theDotProduct;
		bool		isBigHop = false;
		bool		isHopThisFrame = false;
		cveh->GetGravity( &vecSpeedAdd );
		vecSpeedAdd = -vecSpeedAdd;

		cveh->GetMoveSpeed( &vecMoveSpeed );
		theDotProduct = vecMoveSpeed.DotProduct( &vecSpeedAdd );
		if ( theDotProduct < set.vehicle_hop_speed / 2.0f )
			isHopThisFrame = true;
		if ( theDotProduct < -0.1f )
			isBigHop = true;

		for ( ; cveh != NULL; cveh = cveh->GetTowedVehicle() )
		{
			cveh->GetMoveSpeed( &vecMoveSpeed );
			if ( isHopThisFrame )
			{
				if ( isBigHop )
					vecMoveSpeed += ( vecSpeedAdd * (set.vehicle_hop_speed * 2.0f) );
				else
					vecMoveSpeed += ( vecSpeedAdd * (set.vehicle_hop_speed / 2.0f) );
				cveh->SetMoveSpeed( &vecMoveSpeed );
			}

			//set.vehicle_hop_speed;
			if ( !set.trailer_support )
				break;
		}
	}
}

void cheat_handle_vehicle_brake ( struct vehicle_info *info, double time_diff )
{
	traceLastFunc( "cheat_handle_vehicle_brake()" );

	float				speed;
	struct vehicle_info *temp;

	if ( KEY_DOWN(set.key_brake_mod) && !cheat_state->vehicle.air_brake )
	{
		for ( temp = info; temp != NULL; temp = temp->trailer )
		{
			speed = vect3_length( temp->speed );
			vect3_normalize( temp->speed, temp->speed );
			speed -= time_diff * set.brake_mult;

			if ( speed < 0.0f )
				speed = 0.0f;

			if ( vect3_near_zero(temp->speed) )
			{
				vect3_zero( temp->speed );
			}
			else
			{
				if ( temp->vehicle_type == VEHICLE_TYPE_TRAIN )
				{
					if ( temp->m_fTrainSpeed <= 0.05f && temp->m_fTrainSpeed >= -0.05f )
						temp->m_fTrainSpeed = 0.0f;
					else if ( temp->m_fTrainSpeed < 0.0f )
						temp->m_fTrainSpeed += time_diff * set.brake_mult;
					else
						temp->m_fTrainSpeed -= time_diff * set.brake_mult;
				}

				vect3_mult( temp->speed, speed, temp->speed );

				// heh
				int		gonadsMult = 1000;
				float	strifeMult = 0.0000001f;
				int		gonads = rand() % gonadsMult;
				float	strife = (double)gonads * strifeMult;
				if ( strife < strifeMult * gonadsMult / 2 )
					strife -= strifeMult * gonadsMult;
				info->m_SpeedVec.fX += strife;
				gonads = rand() % gonadsMult;
				strife = (double)gonads * strifeMult;
				if ( strife < strifeMult * gonadsMult / 2 )
					strife -= strifeMult * gonadsMult;
				info->m_SpeedVec.fY += strife;
			}

			if ( !set.trailer_support )
				break;
		}
	}
}

void cheat_handle_vehicle_nitro ( struct vehicle_info *info, float time_diff )
{
	traceLastFunc( "cheat_handle_vehicle_nitro()" );

	static uint32_t		timer;
	static int			decelerating;
	static float		speed_off;
	float				pre_speed[3];
	struct vehicle_info *temp;

	if ( KEY_DOWN(set.key_warp_mod) )
		vect3_copy( info->speed, pre_speed );

	if ( KEY_PRESSED(set.key_nitro_mod) )
	{
		speed_off = vect3_length( info->speed );
		decelerating = 0;
		timer = time_get();
	}

	/* "nitro" acceleration mod */
	if ( KEY_DOWN(set.key_nitro_mod) && !vect3_near_zero(info->speed) )
	{
		float	etime = TIME_TO_FLOAT( time_get() - timer ) / set.nitro_accel_time;
		float	speed = set.nitro_high;

		if ( !near_zero(set.nitro_accel_time) )
			etime += 1.0f - ( set.nitro_high - speed_off ) / set.nitro_high;

		if ( etime < 1.0f && !near_zero(set.nitro_accel_time) )
			speed = set.nitro_high * etime;

		for ( temp = info; temp != NULL; temp = temp->trailer )
		{
			if ( !vect3_near_zero(temp->speed) )
			{
				if ( temp->vehicle_type == VEHICLE_TYPE_TRAIN )
				{
					if ( temp->m_fTrainSpeed < 0.0f && temp->m_fTrainSpeed > -set.nitro_high )
						temp->m_fTrainSpeed -= speed;
					else if ( temp->m_fTrainSpeed < set.nitro_high )
						temp->m_fTrainSpeed += speed;
				}

				vect3_normalize( temp->speed, temp->speed );
				vect3_mult( temp->speed, speed, temp->speed );
				if ( vect3_near_zero(temp->speed) )
					vect3_zero( temp->speed );
			}

			if ( !set.trailer_support )
				break;
		}

		// heh
		int		gonadsMult = 1000;
		float	strifeMult = 0.0000001f;
		int		gonads = rand() % gonadsMult;
		float	strife = (double)gonads * strifeMult;
		if ( strife < strifeMult * gonadsMult / 2 )
			strife -= strifeMult * gonadsMult;
		info->m_SpeedVec.fX += strife;
		gonads = rand() % gonadsMult;
		strife = (double)gonads * strifeMult;
		if ( strife < strifeMult * gonadsMult / 2 )
			strife -= strifeMult * gonadsMult;
		info->m_SpeedVec.fY += strife;
	}

	if ( KEY_RELEASED(set.key_nitro_mod) )
	{
		if ( vect3_length(info->speed) > set.nitro_low )
		{
			speed_off = vect3_length( info->speed );
			decelerating = 1;
			timer = time_get();
		}
	}

	if ( decelerating )
	{
		float	speed = set.nitro_low;
		float	etime = TIME_TO_FLOAT( time_get() - timer );

		if ( etime < set.nitro_decel_time )
			speed = speed_off - ( speed_off - speed ) * ( etime / set.nitro_decel_time );
		else
			decelerating = 0;

		for ( temp = info; temp != NULL; temp = temp->trailer )
		{
			if ( vect3_length(temp->speed) > speed )
			{
				vect3_normalize( temp->speed, temp->speed );
				vect3_mult( temp->speed, speed, temp->speed );
			}

			/* // first fix trains before uncommenting this..
			if ( temp->vehicle_type == VEHICLE_TYPE_TRAIN )
			{
				if ( temp->m_fTrainSpeed > speed )
					temp->m_fTrainSpeed = vect3_length( temp->speed );
				else if ( temp->m_fTrainSpeed < -speed )
					temp->m_fTrainSpeed = -vect3_length( temp->speed );
			}
			*/
			if ( !set.trailer_support )
				break;
		}
	}

	/* actual NOS of the game, toggle infinite NOS, w00t! */
	if ( KEY_PRESSED(set.key_nitro) )
	{
		if ( !cheat_state->vehicle.infNOS_toggle_on )
		{
			cheat_state->vehicle.infNOS_toggle_on = true;
			patcher_install( &patch_vehicle_inf_NOS );
			vehicle_addUpgrade( info, 1010 );
		}
		else
		{
			cheat_state->vehicle.infNOS_toggle_on = false;
			patcher_remove( &patch_vehicle_inf_NOS );
			GTAfunc_removeVehicleUpgrade( info, 1010 );
		}
	}
}

//MATRIX4X4
//	VECTOR right			
//		float X [4*0+0]
//		float Y [4*0+1]
//		float Z [4*0+2]		Roll
// DWORD  flags;
// VECTOR up;
//		float X [4*1+0]
//		float Y [4*1+1]
//		float Z [4*1+2]		Pitch
// float  pad_u;
// VECTOR at;				Front
//		float X [4*2+0]
//		float Y [4*2+1]
//		float Z [4*2+2]		Yaw

//
void cheat_handle_vehicle_quick_turn ( struct vehicle_info *info, float time_diff )
{
	traceLastFunc( "cheat_handle_vehicle_quick_turn()" );

	if ( KEY_PRESSED(set.key_quick_turn_180) )
	{
		/* simply invert the X and Y axis.. */
		for ( struct vehicle_info * temp = info; temp != NULL; temp = temp->trailer )
		{
			vect3_invert( &temp->base.matrix[4 * 0], &temp->base.matrix[4 * 0] );
			vect3_invert( &temp->base.matrix[4 * 1], &temp->base.matrix[4 * 1] );
			vect3_invert( temp->speed, temp->speed );
			if ( !cheat_state->vehicle.keep_trailer_attached || !set.trailer_support )
				break;
		}

		if ( info->vehicle_type == VEHICLE_TYPE_TRAIN )
		{
			for ( struct vehicle_info * temp = info; temp != NULL; temp = temp->m_train_next_carriage )
			{
				if ( !g_SAMP )
				{
					temp->m_trainFlags.bDirection ^= 1;
					if ( info->m_train_next_carriage == temp && info->base.model_alt_id == 538 )
					{	//avoid brown streak bug
						if ( temp->m_fDistanceToNextCarriage == 16.5f )
							temp->m_fDistanceToNextCarriage = 20.8718f;
						else
							temp->m_fDistanceToNextCarriage = -16.5f;
					}

					temp->m_fDistanceToNextCarriage *= -1;
				}

				temp->m_fTrainSpeed = -info->m_fTrainSpeed;
			}
		}
	}

	if ( KEY_PRESSED(set.key_quick_turn_left) )
	{
		struct vehicle_info *temp;
		for ( temp = info; temp != NULL; temp = temp->trailer )
		{
			// do new heading
			float		*heading_matrix = temp->base.matrix;
			MATRIX4X4	*heading_matrix4x4 = temp->base.matrixStruct;

			// rotate on z axis
			CVehicle	*cveh = pGameInterface->GetPools()->GetVehicle( (DWORD *)temp );
			CVector		posUnder = cheat_vehicle_getPositionUnder( cveh );
			float		heading_vector_zrotate[3] = { posUnder.fX, posUnder.fY, posUnder.fZ };
			float		heading_theta = M_PI / 2.0f;
			vect3_normalize( heading_vector_zrotate, heading_vector_zrotate );
			matrix_vect3_rotate( heading_matrix, heading_vector_zrotate, heading_theta, heading_matrix );

			// do new speed
			if ( !temp->m_SpeedVec.IsNearZero() )
			{
				temp->m_SpeedVec.CrossProduct( &posUnder );
			}

			if ( !cheat_state->vehicle.keep_trailer_attached || !set.trailer_support )
				break;
		}
	}

	if ( KEY_PRESSED(set.key_quick_turn_right) )
	{
		struct vehicle_info *temp;
		for ( temp = info; temp != NULL; temp = temp->trailer )
		{
			// do new heading
			float		*heading_matrix = temp->base.matrix;
			MATRIX4X4	*heading_matrix4x4 = temp->base.matrixStruct;

			// rotate on z axis
			CVehicle	*cveh = pGameInterface->GetPools()->GetVehicle( (DWORD *)temp );
			CVector		posUnder = cheat_vehicle_getPositionUnder( cveh );
			posUnder = -posUnder;

			float	heading_zvectrotate[4] = { posUnder.fX, posUnder.fY, posUnder.fZ };
			float	heading_theta = M_PI / 2.0f;
			vect3_normalize( heading_zvectrotate, heading_zvectrotate );
			matrix_vect3_rotate( heading_matrix, heading_zvectrotate, heading_theta, heading_matrix );

			// do new speed
			if ( !temp->m_SpeedVec.IsNearZero() )
			{
				temp->m_SpeedVec.CrossProduct( &posUnder );
			}

			if ( !cheat_state->vehicle.keep_trailer_attached || !set.trailer_support )
				break;
		}
	}
}

void cheat_handle_vehicle_protection ( struct vehicle_info *info, float time_diff )
{
	traceLastFunc( "cheat_handle_vehicle_protection()" );

	static float		last_spin[3];
	struct vehicle_info *temp;

	if ( KEY_PRESSED(set.key_protection) )
		cheat_state->vehicle.protection ^= 1;

	if ( KEY_DOWN(set.key_warp_mod) )
		return;

	if ( cheat_state->vehicle.protection )
	{
		for ( temp = info; temp != NULL; temp = temp->trailer )
		{
			if ( !vect3_near_zero(temp->spin) && vect3_length(temp->spin) > set.protection_spin_cap )
			{
				vect3_normalize( temp->spin, temp->spin );
				vect3_mult( temp->spin, set.protection_spin_cap, temp->spin );
			}

			if ( !vect3_near_zero(temp->speed) && vect3_length(temp->speed) > set.protection_speed_cap )
			{
				vect3_normalize( temp->speed, temp->speed );
				vect3_mult( temp->speed, set.protection_speed_cap, temp->speed );
			}

			if ( temp->base.matrix[4 * 2 + 2] < 0.0f )
			{
				if ( !vect3_near_zero(last_spin) && !vect3_near_zero(temp->spin) && vect3_length(temp->spin) < 0.10f )
				{
					vect3_normalize( temp->spin, temp->spin );
					vect3_mult( temp->spin, 0.10f, temp->spin );
				}
			}
			else if ( temp == info )	/* only copy the first vehicle's spin info */
			{
				vect3_copy( temp->spin, last_spin );
			}

			vehicle_prevent_below_height( temp, set.protection_min_height );

			if ( !set.trailer_support )
				break;
		}
	}
}

void cheat_handle_vehicle_engine ( struct vehicle_info *vehicle_info, float time_diff )
{
	if ( vehicle_info == NULL )
		return;

	traceLastFunc( "cheat_handle_vehicle_engine()" );

	struct vehicle_info *veh_self = vehicle_info_get( VEHICLE_SELF, 0 );
	if ( veh_self == NULL )
		return;

	if ( KEY_PRESSED(set.key_engine) )
	{
		if ( veh_self->m_nVehicleFlags.bEngineOn )
		{
			veh_self->m_nVehicleFlags.bEngineOn = 0;
			cheat_state->vehicle.is_engine_on = 0;
		}
		else
		{
			veh_self->m_nVehicleFlags.bEngineOn = 1;
			cheat_state->vehicle.is_engine_on = 1;
		}
	}
}

void cheat_handle_vehicle_brakedance ( struct vehicle_info *vehicle_info, float time_diff )
{
	if ( cheat_state->vehicle.air_brake )
		return;
	if ( cheat_state->vehicle.stick )
		return;

	traceLastFunc( "cheat_handle_vehicle_brakedance()" );

	static float	velpos, velneg;

	if ( KEY_PRESSED(set.key_brkd_toggle) )
	{
		cheat_state->vehicle.brkdance ^= 1;
	}

	if ( cheat_state->vehicle.brkdance )
	{
		// we should probably actually be driving the vehicle
		struct actor_info	*actor = actor_info_get( ACTOR_SELF, 0 );
		if ( actor == NULL )
			return; // we're not an actor? lulz
		if ( actor->state != ACTOR_STATE_DRIVING )
			return; // we're not driving?
		if ( actor->vehicle->passengers[1] == actor )
			return; // we're not passenger in an airplane?
		int		iVehicleID = ScriptCarId( vehicle_info );

		float	fTimeStep = *(float *)0xB7CB5C;

		velpos = set.brkdance_velocity * fTimeStep;
		velneg = -set.brkdance_velocity * fTimeStep;

		if ( KEY_DOWN(set.key_brkd_forward) )
		{
			ScriptCommand( &apply_momentum_in_direction_XYZ, iVehicleID, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f );
			ScriptCommand( &apply_rotory_pulse_about_an_axis_XYZ, iVehicleID, velneg, 0.0f, 0.0f );
		}
		else if ( KEY_DOWN(set.key_brkd_backward) )
		{
			ScriptCommand( &apply_momentum_in_direction_XYZ, iVehicleID, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f );
			ScriptCommand( &apply_rotory_pulse_about_an_axis_XYZ, iVehicleID, velpos, 0.0f, 0.0f );
		}
		else if ( KEY_DOWN(set.key_brkd_right) )
		{
			ScriptCommand( &apply_momentum_in_direction_XYZ, iVehicleID, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f );
			ScriptCommand( &apply_rotory_pulse_about_an_axis_XYZ, iVehicleID, 0.0f, velpos, 0.0f );
		}
		else if ( KEY_DOWN(set.key_brkd_left) )
		{
			ScriptCommand( &apply_momentum_in_direction_XYZ, iVehicleID, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f );
			ScriptCommand( &apply_rotory_pulse_about_an_axis_XYZ, iVehicleID, 0.0f, velneg, 0.0f );
		}
		else if ( KEY_DOWN(set.key_brkd_rightward) )
		{
			ScriptCommand( &apply_momentum_in_direction_XYZ, iVehicleID, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f );
			ScriptCommand( &apply_rotory_pulse_about_an_axis_XYZ, iVehicleID, 0.0f, 0.0f, velneg );
		}
		else if ( KEY_DOWN(set.key_brkd_leftward) )
		{
			ScriptCommand( &apply_momentum_in_direction_XYZ, iVehicleID, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f );
			ScriptCommand( &apply_rotory_pulse_about_an_axis_XYZ, iVehicleID, 0.0f, 0.0f, velpos );
		}
		else
			iVehicleID = -1;

		if ( set.trailer_support && iVehicleID != -1 )
		{
			for ( struct vehicle_info * temp = vehicle_info; temp != NULL; temp = temp->trailer )
			{
				vect3_copy( vehicle_info->spin, temp->spin );
			}
		}
	}
}

void cheat_handle_vehicle_blinking_carlights ( struct vehicle_info *vinfo, float time_diff )
{
	traceLastFunc( "cheat_handle_blinking_carlights()" );

	int class_id = gta_vehicle_get_by_id( vinfo->base.model_alt_id )->class_id;
	if ( vinfo->vehicle_type != VEHICLE_TYPE_CAR
	 ||	 class_id == VEHICLE_CLASS_TRAILER
	 ||	 class_id == VEHICLE_CLASS_AIRPLANE
	 ||	 class_id == VEHICLE_CLASS_HELI
	 ||	 class_id == VEHICLE_CLASS_BIKE ) return;

	if ( KEY_PRESSED(set.key_blinking_car_lights) )
	{
		// reset lights damage
		vinfo->lights_status_rear = false;
		vinfo->lights_status_frontLeft = false;
		vinfo->lights_status_frontRight = false;

		// increment lights state
		cheat_state->vehicle.blinking_carlights_state++;

		// set it back down to zero if over the max
		if ( cheat_state->vehicle.blinking_carlights_state > 3 )
			cheat_state->vehicle.blinking_carlights_state = 0;

		// reset turn state
		cheat_state->vehicle.blinking_carlights_turnstate = 0;
	}

	// return if lights state is 0/off
	if ( !cheat_state->vehicle.blinking_carlights_state )
		return;

	// setup variables for next blinking lights modes
	struct actor_info	*actor_self = actor_info_get( ACTOR_SELF, ACTOR_ALIVE );
	if ( vinfo->passengers[0] != actor_self )
		return;

	// enables car lights any time of day, not synced
	// we need a better way to do this that can be disabled
	//ScriptCommand( &set_car_lights, ScriptCarId(vinfo), 2 );
	//
	// switch is faster than multiple if/else statements
	switch ( cheat_state->vehicle.blinking_carlights_state )
	{
	case 1:
		// turn lights
		// blink rate limiter
		if ( (GetTickCount() - 250) < cheat_state->vehicle.blinking_carlights_lastblink )
			return;
		if ( cheat_state->vehicle.blinking_carlights_turnstate == 0 )
		{
			if ( vinfo->steer_angles[0] <= -0.01f )
			{
				vinfo->lights_status_frontLeft = false;
				vinfo->lights_status_frontRight = true;
			}
			else if ( vinfo->steer_angles[0] >= 0.01f )
			{
				vinfo->lights_status_frontLeft = true;
				vinfo->lights_status_frontRight = false;
			}

			cheat_state->vehicle.blinking_carlights_turnstate = 1;
		}
		else
		{
			vinfo->lights_status_frontLeft = false;
			vinfo->lights_status_frontRight = false;
			cheat_state->vehicle.blinking_carlights_turnstate = 0;
		}
		break;

	case 2:
		// police style lights
		// blink rate limiter
		if ( (GetTickCount() - 150) < cheat_state->vehicle.blinking_carlights_lastblink )
			return;

		// rear lights
		if ( vinfo->lights_status_rear )
			vinfo->lights_status_rear = false;
		else
			vinfo->lights_status_rear = true;

		// front lights
		switch ( cheat_state->vehicle.blinking_carlights_turnstate )
		{
		case 0:
			vinfo->lights_status_frontLeft = true;
			vinfo->lights_status_frontRight = false;
			break;

		case 1:
			vinfo->lights_status_frontLeft = false;
			vinfo->lights_status_frontRight = true;
			break;

		case 2:
			vinfo->lights_status_frontLeft = true;
			vinfo->lights_status_frontRight = true;
			break;

		case 3:
			vinfo->lights_status_frontLeft = false;
			vinfo->lights_status_frontRight = true;
			break;

		case 4:
			vinfo->lights_status_frontLeft = true;
			vinfo->lights_status_frontRight = false;
			break;

		case 5:
			vinfo->lights_status_frontLeft = true;
			vinfo->lights_status_frontRight = true;
			break;
		}

		// increment/reset turnstate
		cheat_state->vehicle.blinking_carlights_turnstate++;
		if ( cheat_state->vehicle.blinking_carlights_turnstate > 5 )
			cheat_state->vehicle.blinking_carlights_turnstate = 0;
		break;

	case 3:
		// stroboscope
		// blink rate limiter
		if ( (GetTickCount() - 25) < cheat_state->vehicle.blinking_carlights_lastblink )
			return;

		// rear lights
		if ( vinfo->lights_status_rear )
		{
			vinfo->lights_status_rear = false;
			vinfo->lights_status_frontLeft = true;
			vinfo->lights_status_frontRight = true;
		}
		else
		{
			vinfo->lights_status_rear = true;
			vinfo->lights_status_frontLeft = false;
			vinfo->lights_status_frontRight = false;
		}
		break;
	}

	// reset counter
	cheat_state->vehicle.blinking_carlights_lastblink = GetTickCount();
}

void cheat_handle_vehicle_fly ( struct vehicle_info *vehicle_info, float time_diff )
{
	traceLastFunc( "cheat_handle_vehicle_fly()" );

	static bool					needRestorePphys = false;
	static float				plane_orig_data[3];			// pitch, roll, circle
	static struct vehicle_info	*last_plane;

	// getting passed a NULL pointer from cheat_panic, so we can remove the patch and reapply airplane physics
	if ( vehicle_info == NULL )
	{
		if ( !cheat_state->_generic.cheat_panic_enabled )
			return;

		if ( patch_NotAPlane.installed )
			patcher_remove( &patch_NotAPlane );

		struct vehicle_info *veh_self = vehicle_info_get( VEHICLE_SELF, NULL );

		if ( veh_self == NULL )
			return;

		if ( gta_vehicle_get_by_id(veh_self->base.model_alt_id)->class_id == VEHICLE_CLASS_AIRPLANE )
		{
			veh_self->pFlyData->pitch = plane_orig_data[0];
			veh_self->pFlyData->roll_lr = plane_orig_data[1];
			veh_self->pFlyData->circleAround = plane_orig_data[2];
			needRestorePphys = false;
		}

		return;
	}

	// this should never happen
	if ( pGameInterface == NULL )
		return;

	if ( KEY_PRESSED(set.key_flyMode_change) )
		set.fly_heliMode = !set.fly_heliMode;

	if ( KEY_PRESSED(set.key_fly_vehicle) )
		cheat_state->vehicle.fly ^= 1;

	// ignore hydra, RC Baron and RC Goblin (they seem to use some special functions to fly)
	if ( vehicle_info->base.model_alt_id == 520 || vehicle_info->base.model_alt_id == 464
		|| vehicle_info->base.model_alt_id == 501 )
	{
		if ( patch_NotAPlane.installed )
			patcher_remove( &patch_NotAPlane );
		return;
	}

	if ( patch_NotAPlane.installed && !cheat_state->vehicle.fly )
		patcher_remove( &patch_NotAPlane );

	int class_id = gta_vehicle_get_by_id( vehicle_info->base.model_alt_id )->class_id;
	if ( cheat_state->vehicle.fly )
	{
		if ( class_id == VEHICLE_CLASS_HELI )
			return;

		if ( class_id == VEHICLE_CLASS_AIRPLANE )
		{
			if ( last_plane != vehicle_info )
			{
				plane_orig_data[0] = vehicle_info->pFlyData->pitch;
				plane_orig_data[1] = vehicle_info->pFlyData->roll_lr;
				plane_orig_data[2] = vehicle_info->pFlyData->circleAround;
				last_plane = vehicle_info;
			}

			if ( !patch_NotAPlane.installed )
				patcher_install( &patch_NotAPlane );
			needRestorePphys = true;
		}
		else if ( patch_NotAPlane.installed )
		{
			patcher_remove( &patch_NotAPlane );
		}

		struct vehicle_info *temp;
		DWORD				func = 0x006D85F0;
		for ( temp = vehicle_info; temp != NULL; temp = temp->trailer )
		{
			DWORD	mecar = ( DWORD ) temp;
			class_id = gta_vehicle_get_by_id( temp->base.model_alt_id )->class_id;

			// fly physics heli Mode / Bike
			if ( set.fly_heliMode || class_id == VEHICLE_CLASS_BIKE )
			{
				temp->pFlyData->pitch = 0.0035f;

				if ( class_id == VEHICLE_CLASS_BIKE )
				{
					temp->pFlyData->roll_lr = -0.01f; // rolling isn't working with motorized bikes yet
					temp->pFlyData->circleAround = -0.0006f;
				}
				else
				{
					temp->pFlyData->roll_lr = -0.004f;
					temp->pFlyData->circleAround = -0.0003f;
				}
			}

			// fly physics plane Mode
			else
			{
				// use original physics for planes
				if ( class_id == VEHICLE_CLASS_AIRPLANE )
				{
					vehicle_info->pFlyData->pitch = plane_orig_data[0];
					vehicle_info->pFlyData->roll_lr = plane_orig_data[1];
					vehicle_info->pFlyData->circleAround = plane_orig_data[2];
				}
				else
				{
					temp->pFlyData->pitch = 0.0005f;
					temp->pFlyData->roll_lr = 0.005f;
					temp->pFlyData->circleAround = -0.001f;
				}
			}

			//  steering  //
			float	one = 0.9997f;
			float	min = -0.9997f;
			if ( *(uint8_t *) (GTA_KEYS + 0x1C) == 0xFF )	//accel
			{
				__asm push min
			}
			else if ( *(uint8_t *) (GTA_KEYS + 0x20) == 0xFF )	//brake
			{
				__asm push one
			}
			else
			{
				__asm push 0
			}

			if ( *(uint8_t *) (GTA_KEYS + 0x1) == 0xFF )		//left
			{
				__asm push min
			}
			else if ( *(uint8_t *) (GTA_KEYS + 0x0) == 0x80 )	//right
			{
				__asm push one
			}
			else
			{
				__asm push 0
			}

			if ( *(uint8_t *) (GTA_KEYS + 0x3) == 0xFF )		//steer forward
			{
				__asm push min
			}
			else if ( *(uint8_t *) (GTA_KEYS + 0x2) == 0x80 )	//steer down
			{
				__asm push one
			}
			else
			{
				__asm push 0
			}

			if ( *(uint8_t *) (GTA_KEYS + 0xE) == 0xFF )		//look left
			{
				__asm push one
			}
			else if ( *(uint8_t *) (GTA_KEYS + 0xA) == 0xFF )	//Look right
			{
				__asm push min
			}
			else
			{
				__asm push 0
			}
			//   steering end    //

			// 1 fast plane, 2 heli, 6 heli, 8 airbreak alike
			if ( set.fly_heliMode )
			{
				__asm push 6
			}
			else
			{
				__asm push 1
			}

			__asm mov ecx, mecar
			__asm call func

			// no trailer support
			if ( !set.trailer_support )
				return;

			// sorta fix trailer spin
			if ( temp != vehicle_info )
				vect3_copy( vehicle_info->spin, temp->spin );
		}
	}

	// change airplane physics back to normal
	else if ( class_id == VEHICLE_CLASS_AIRPLANE && needRestorePphys )
	{
		vehicle_info->pFlyData->pitch = plane_orig_data[0];
		vehicle_info->pFlyData->roll_lr = plane_orig_data[1];
		vehicle_info->pFlyData->circleAround = plane_orig_data[2];
		needRestorePphys = false;
	}
}

void cheat_handle_vehicle_keepTrailer ( struct vehicle_info *vinfo, float time_diff )
{
	traceLastFunc( "cheat_handle_vehicle_keepTrailer()" );

	// toggle state from key press
	if ( KEY_PRESSED(set.key_keep_trailer) )
		cheat_state->vehicle.keep_trailer_attached ^= 1;

	// return if disabled or not driving
	if ( !cheat_state->vehicle.keep_trailer_attached )
		return;

	struct actor_info	*self = actor_info_get( ACTOR_SELF, ACTOR_ALIVE );
	if ( self == NULL || vinfo->passengers[0] != self )
		return;

	// static vars to store trailer info
	static struct vehicle_info	*myveh_old;
	static struct vehicle_info	*mytrailer_old;

	if ( vinfo == myveh_old )
	{
		if ( vinfo->trailer != NULL )
		{
			mytrailer_old = vinfo->trailer;
			return;
		}
		else if ( mytrailer_old != NULL )
		{
			DWORD	trailer_id = ScriptCarId( mytrailer_old );
			if ( trailer_id == NULL )
				return;

			CVector distance = vinfo->base.m_CMatrix->vPos - mytrailer_old->base.m_CMatrix->vPos;
			if ( distance.Length() <= 40.0f )
			{
				// new variables to be used with R* classes
				CVehicle	*cveh, *cvehtrailer;
				CVector		cvehGrav, rotationAxis, vecVehTowBarPos, vecTrailerHitchPos, vecTrailerPos;
				CVector		vZero ( 0.0f, 0.0f, 0.0f );
				CMatrix		cvehMatrix;
				float		theta;

				if ( !cheat_state->vehicle.air_brake )
				{
					// unflip
					// get this vehicle, gravity, and matrix
					cveh = pPools->GetVehicle( (DWORD *)vinfo );
					cveh->GetGravity( &cvehGrav );
					cveh->GetMatrix( &cvehMatrix );

					// get trailer
					cvehtrailer = pPools->GetVehicle( (DWORD *)mytrailer_old );

					// get "down" from vehicle model
					rotationAxis = cheat_vehicle_getPositionUnder( cveh );

					// normalize our vectors
					cvehGrav.Normalize();
					rotationAxis.Normalize();

					// axis and rotation for gravity
					theta = acos( rotationAxis.DotProduct(&cvehGrav) );
					if ( theta > FLOAT_EPSILON )
					{
						rotationAxis.CrossProduct( &cvehGrav );
						rotationAxis.Normalize();
						rotationAxis.ZeroNearZero();
						cvehMatrix = cvehMatrix.Rotate( &rotationAxis, -theta );

						// set the new matrix
						cveh->SetMatrix( &cvehMatrix );
					}

					// re-attach trailer
					cvehtrailer->SetTowLink( cveh );

					// fix trailer orientation
					rotationAxis = cheat_vehicle_getPositionUnder( cvehtrailer );
					cvehtrailer->GetMatrix( &cvehMatrix );
					theta = acos( rotationAxis.DotProduct(&cvehGrav) );
					if ( theta > FLOAT_EPSILON )
					{
						rotationAxis.CrossProduct( &cvehGrav );
						rotationAxis.Normalize();
						rotationAxis.ZeroNearZero();
						cvehMatrix = cvehMatrix.Rotate( &rotationAxis, -theta );
						cvehMatrix = cvehMatrix.Rotate( &cvehMatrix.vRight, 0.15f );
						cvehtrailer->SetMatrix( &cvehMatrix );
					}

					// fix trailer position
					cveh->GetTowBarPos( &vecVehTowBarPos );
					cvehtrailer->GetTowHitchPos( &vecTrailerHitchPos );
					vecTrailerPos = cvehtrailer->GetInterface()->Placeable.matrix->vPos;
					vecTrailerPos += vecVehTowBarPos - vecTrailerHitchPos;
					cvehtrailer->SetPosition( &vecTrailerPos );

					// prevent collision from altering speed/spin
					//cheat_state->_generic.nocols_enabled = 1;
					//cheat_state->_generic.nocols_change_tick = GetTickCount();
					cveh->GetVehicleInterface()->vecVelocityCollision = &vZero;
					cveh->GetVehicleInterface()->vecSpinCollision = &vZero;
					cvehtrailer->GetVehicleInterface()->vecVelocityCollision = &vZero;
					cvehtrailer->GetVehicleInterface()->vecSpinCollision = &vZero;
				}
				else
				{
					// get vehicle pointers
					cveh = pPools->GetVehicle( (DWORD *)vinfo );
					cvehtrailer = pPools->GetVehicle( (DWORD *)mytrailer_old );

					// re-attach trailer
					cvehtrailer->SetTowLink( cveh );

					// prevent collision from altering speed/spin
					cheat_state->_generic.nocols_enabled = 1;
					cheat_state->_generic.nocols_change_tick = GetTickCount();
				}
			}
			else
			{
				mytrailer_old = NULL;
			}
		}
	}
	else if ( vinfo->trailer != NULL )
	{
		myveh_old = vinfo;
		mytrailer_old = vinfo->trailer;
	}
	else
	{
		myveh_old = vinfo;
		mytrailer_old = NULL;
	}
}

// remove vehicle_info variable + move into cheat_generic (?)
void cheat_handle_vehicle_fast_exit ( struct vehicle_info *vehicle_info, float time_diff )
{
	traceLastFunc( "cheat_handle_fast_exit()" );

	if ( KEY_PRESSED(set.key_fast_exit) )
	{
		float				*coord =
			( cheat_state->state == CHEAT_STATE_VEHICLE ) ? cheat_state->vehicle.coords : cheat_state->actor.coords;
		struct actor_info	*self = actor_info_get( ACTOR_SELF, ACTOR_ALIVE );
		ScriptCommand( &remove_actor_from_car_and_put_at, ScriptActorId(self), coord[0], coord[1] + 2.0f, coord[2] );
	}
}

void cheat_handle_vehicle_repair_car ( struct vehicle_info *vehicle_info, float time_diff )
{
	traceLastFunc( "cheat_handle_repair_car()" );

	if ( KEY_PRESSED(set.key_repair_car) )
	{
		// get info
		struct vehicle_info *veh_self = vehicle_info_get( VEHICLE_SELF, 0 );
		if ( veh_self == NULL )
			return;

		struct actor_info	*self = actor_info_get( ACTOR_SELF, 0 );
		if ( self == NULL )
			return;

		// make sure we are driving our own vehicle
		if ( self->state == ACTOR_STATE_DRIVING && self->vehicle->passengers[0] == self )
		{
			struct vehicle_info *temp;

			//fix the vehicle
			for ( temp = vehicle_info; temp != NULL; temp = temp->trailer )
			{
				int iVehicleID = ScriptCarId( temp );
				ScriptCommand( &repair_car, iVehicleID );

				if ( !set.trailer_support )
					break;
			}
		}
	}
}

// ---------------------------------------------------------
// ---------------------------------------------------------

/*
CMatrix
	vRight = CVector ( 1.0f, 0.0f, 0.0f );
	vFront = CVector ( 0.0f, 1.0f, 0.0f );
	vUp    = CVector ( 0.0f, 0.0f, 1.0f );
	vPos   = CVector ( 0.0f, 0.0f, 0.0f );
MATRIX4X4
	VECTOR right;	// vRight
	VECTOR up;		// vFront
	VECTOR at;		// vUp
	VECTOR pos;		// vPos
*/
CVector cheat_vehicle_getPositionUnder ( CVehicle *cveh )
{
	traceLastFunc( "cheat_vehicle_getPositionUnder()" );

	CVector offsetVector;
	CMatrix matrix;
	cveh->GetMatrix( &matrix );

	CMatrix_Padded	matrixPad ( matrix );
	offsetVector.fX = 0 * matrixPad.vRight.fX + 0 * matrixPad.vFront.fX - 1 * matrixPad.vUp.fX;
	offsetVector.fY = 0 * matrixPad.vRight.fY + 0 * matrixPad.vFront.fY - 1 * matrixPad.vUp.fY;
	offsetVector.fZ = 0 * matrixPad.vRight.fZ + 0 * matrixPad.vFront.fZ - 1 * matrixPad.vUp.fZ;
	return offsetVector;
}

#ifdef M0D_DEV

/*
D3DXVECTOR3 vecGravColOrigin, vecGravColTarget, vecGravTargetNorm;
CVector temp_vecGravTargetNorm;
*/
#endif
void cheat_vehicle_setGravity ( CVehicle *cveh, CVector pvecGravity )
{
	traceLastFunc( "cheat_vehicle_setGravity()" );

	// set the d-dang gravity
	cveh->SetGravity( &pvecGravity );

	//5:08:18 PM lol cool
}

struct patch_set	*patchBikeFalloff_set = NULL;
bool				m_SpiderWheels_falloffFound = false;
bool				m_SpiderWheels_falloffEnabled = false;
bool init_patchBikeFalloff ( void )
{
	traceLastFunc( "init_patchBikeFalloff()" );

	if ( !m_SpiderWheels_falloffFound )
	{
		int patchBikeFalloff_ID = GTAPatchIDFinder( 0x004BA3B9 );
		if ( patchBikeFalloff_ID != -1 )
		{
			patchBikeFalloff_set = &set.patch[patchBikeFalloff_ID];
			m_SpiderWheels_falloffFound = true;
		}
		else
		{
			Log( "Couldn't init_patchBikeFalloff." );
			Log( "You may fall off bikes while using SpiderWheels." );
			Log( "Put the 'Anti bike fall off' patch back into your INI to fix this problem." );
		}
	}

	return m_SpiderWheels_falloffFound;
}

void cheat_handle_vehicle_spiderWheels ( struct vehicle_info *vinfo, float time_diff )
{
	traceLastFunc( "cheat_handle_spiderWheels()" );

	// 3:07:16 PM how you going
	if ( KEY_PRESSED(set.key_spiderwheels) )
	{
		// init variables used to toggle patch
		init_patchBikeFalloff();

		// toggle the d-dang spiderz
		cheat_state->vehicle.spiderWheels_on ^= 1;
	}

	if ( cheat_state->vehicle.spiderWheels_on )
	{
		// init SpiderWheels
		if ( !cheat_state->vehicle.spiderWheels_Enabled )
		{
			// install anti bike falloff patch if needed
			if ( m_SpiderWheels_falloffFound && !m_SpiderWheels_falloffEnabled && !patchBikeFalloff_set->installed )
			{
				patcher_install( patchBikeFalloff_set );
				m_SpiderWheels_falloffEnabled = true;
			}

			// set spider wheels enabled
			cheat_state->vehicle.spiderWheels_Enabled = true;
		}

		// loop through the vehicle and any trailers
		for ( vehicle_info * temp = vinfo; temp != NULL; temp = temp->trailer )
		{
			// get CVehicle
			CVehicle			*cveh = pGameInterface->GetPools()->GetVehicle( (DWORD *)temp );

			// update spider wheels
			CVector				offsetVector = cheat_vehicle_getPositionUnder( cveh );

			// setup variables
			CVector				vecOrigin, vecTarget;
			CColPoint			*pCollision = NULL;
			CEntitySAInterface	*pCollisionEntity = NULL;
			int					checkDistanceMeters = 20;

			// get CEntitySAInterface pointer
			CEntitySAInterface	*cveh_interface = cveh->GetInterface();

			// origin = our vehicle
			vecOrigin = cveh_interface->Placeable.matrix->vPos;
			vecTarget = offsetVector * checkDistanceMeters;
			vecTarget = vecTarget + vecOrigin;

			// check for collision
			bool	bCollision = GTAfunc_ProcessLineOfSight( &vecOrigin, &vecTarget, &pCollision, &pCollisionEntity, 1,
															 0, 0, 1, 1, 0, 0, 0 );

			if ( bCollision )
			{
				// set altered gravity vector
				float	fTimeStep = *(float *)0xB7CB5C;
				CVector colGravTemp = -pCollision->GetInterface()->Normal;
				CVector vehGravTemp;	//= cheat_state->vehicle.gravityVector;
				cveh->GetGravity( &vehGravTemp );

				CVector newRotVector;
				newRotVector = colGravTemp - vehGravTemp;
				newRotVector *= 0.05f * fTimeStep;
				offsetVector = vehGravTemp + newRotVector;

				// destroy the collision object
				pCollision->Destroy();
			}
			else
			{
				// set normal gravity vector
				float	fTimeStep = *(float *)0xB7CB5C;
				CVector colGravTemp ( 0.0, 0.0, -1.0 );
				CVector vehGravTemp;
				cveh->GetGravity( &vehGravTemp );

				CVector newRotVector;
				newRotVector = colGravTemp - vehGravTemp;
				newRotVector *= 0.05f * fTimeStep;
				offsetVector = vehGravTemp + newRotVector;
			}

			// set the gravity/camera
			offsetVector.ZeroNearZero();
			cheat_vehicle_setGravity( cveh, offsetVector );
		}
	}
	else if ( cheat_state->vehicle.spiderWheels_Enabled )
	{
		// remove anti bike falloff patch if needed
		if ( m_SpiderWheels_falloffFound && m_SpiderWheels_falloffEnabled )
		{
			if ( patchBikeFalloff_set->installed || patchBikeFalloff_set->failed )
			{
				patcher_remove( patchBikeFalloff_set );
			}

			m_SpiderWheels_falloffEnabled = false;
		}

		// loop through the vehicle and any trailers
		for ( vehicle_info * temp = vinfo; temp != NULL; temp = temp->trailer )
		{
			// get vehicle
			CVehicle	*cveh = pGameInterface->GetPools()->GetVehicle( (DWORD *)temp );

			// disable spider wheels with normal gravity vector
			CVector		offsetVector ( 0.0, 0.0, -1.0 );

			// set the gravity/camera
			cheat_vehicle_setGravity( cveh, offsetVector );
		}

		// set spider wheels disabled
		cheat_state->vehicle.spiderWheels_Enabled = false;
	}
}
